﻿/*
 * * * Compile_AHK SETTINGS BEGIN * * *

[AHK2EXE]
Exe_File=%In_Dir%\esp2dsd_wrapper.exe
No_UPX=1
[VERSION]
Set_Version_Info=1
File_Description=esp2dsd の GUI ラッパー
File_Version=0.2.0.0
Inc_File_Version=0
Legal_Copyright=2025 (c) Tondabayashi
Product_Version=0.2.0.0

* * * Compile_AHK SETTINGS END * * *
*/

#Requires Autohotkey v2.0
#SingleInstance force
;#Warn VarUnset, Off
SendMode("Input")
FileEncoding("UTF-8-RAW")
SetWorkingDir(A_ScriptDir)
#Include ./lib/JXON-ah2/_JXON.ahk ;https://github.com/TheArkive/JXON_ahk2

AppName := "esp2dsd Wrapper"
AppVersion := "0.2.0"

exe := ".\esp2dsd.exe"
exeDirOriginal := ".\Original"
exeDirTranslated := ".\Translated"
exeDirOutput := ".\Output"

SplitPath(A_ScriptName, , , &scriptNameExt, &scriptNameNoExt)
iniPath := ".\" . scriptNameNoExt . ".ini"

Tray := A_TrayMenu
Tray.Delete()
TrayStd := Menu()
TrayStd.AddStandard()
Tray.Add(AppName . " " . AppVersion, TrayItemDummy)
Tray.Disable(AppName . " " . AppVersion)
Tray.Add()
if(scriptNameExt != "exe"){
  Tray.Add("AutoHotkey(&A)", TrayStd)
  TraySetIcon(".\" . scriptNameNoExt . ".ico",, true)
}
Tray.Add("終了(&Q)", TrayItemQuit)
A_IconTip := AppName . " " . AppVersion

attrExe := FileExist(exe)
if( ! attrExe ) {
  MsgBox("esp2dsd.exe と同じフォルダに配置してください。")
  ExitApp()
}

;iniの読み込み
readConfig()

myGui := Gui()
myGui.OnEvent("Close", GuiClose)
myGui.OnEvent("Escape", GuiClose)
myGui.OnEvent("DropFiles", GuiDropFiles)
myGui.Opt("+AlwaysOnTop")
myGui.Add("text", , "このウィンドウに、xTranslatorでリネームされた`n翻訳前のesp (PluginName.esp.2025_00_00_00_00_00) と`n翻訳後のesp (PluginName.esp)をドラッグ＆ドロップしてください。`n`n処理が完了すると翻訳後のespと同じフォルダにSKSEフォルダが作成され、`nその中にDSD用のjsonが生成されます。`n(本スクリプトと同じ場所にあるiniを編集すれば出力先を変更できます)")
myGui.Show("X0 Y0")
return

TrayItemQuit(*) { 
  ExitApp()
}
TrayItemDummy(*) { 
  return
}

GuiClose(thisGui) { 
  ExitApp()
}
GuiDropFiles(thisGui, Ctrl, FileArray, *)
{
  global
  arrayDropped := Array()
  arrayOriginal := Array()
  arrayTranslated := Array()
  countResult := 0
  strResult0 := "計"
  strResult1 := "ファイルの処理が完了しました。`n保存されたJSONファイルを確認してください。`n-----------"
  
  ;ドロップされたファイルを配列に起こす
  for fileDropped in FileArray
  {
    ;フォルダではないかチェック
    attrDropped := FileExist(fileDropped)
    isDir := InStr(attrDropped, "D") 
    if( isDir > 0 ) {
      continue
    }
    arrayDropped.Push(fileDropped)
  }
  
  ;ドロップされたファイルから翻訳後のもの（～.esp）を抽出する
  for i, fileDropped in arrayDropped
  {
    
    ;拡張子のチェック
    SplitPath(fileDropped, , , &fileNameExt)
    fileNameExt := StrLower(fileNameExt)
    if( fileNameExt != "esp" && fileNameExt != "esm" && fileNameExt != "esl" ) {
      continue
    }
    ;翻訳前ではないかチェック
    isOriginal := RegExMatch(fileDropped, "\.\d\d\d\d_\d\d_\d\d_\d\d_\d\d_\d\d$")
    if( isOriginal > 0 ) {
      continue
    }
    ;配列に格納
    arrayTranslated.Push(fileDropped)
    arrayOriginal.Push("")
  }
  
  ;翻訳後のファイルに対応するオリジナル（～.esp.202X_～）を探す
  for i, fileTranslated in arrayTranslated
  {
    SplitPath(fileTranslated, &fileNameTranslated)
    indexTranslated := i
    
    ;全ファイルを精査して対応するものだけを配列に格納する
    for j, fileDropped in arrayDropped
    {
      isOriginal := RegExMatch(fileDropped, "\.\d\d\d\d_\d\d_\d\d_\d\d_\d\d_\d\d$")
      
      if( isOriginal > 0 ) {
        SplitPath(fileDropped, &fileNameDropped, , , &fileNameDroppedNoExt)
        if( fileNameDroppedNoExt == fileNameTranslated ) {
        
          ;既に対応するファイルが存在する場合、過去のものを優先する
          fileSaved := ""
          Try fileSaved := arrayOriginal[indexTranslated]
          if( IsSet(fileSaved) && fileSaved != "" ) {
            RegExMatch(fileSaved, "\.\d\d\d\d_\d\d_\d\d_\d\d_\d\d_\d\d$", &stampSaved)
            RegExMatch(fileDropped, "\.\d\d\d\d_\d\d_\d\d_\d\d_\d\d_\d\d$", &stampDropped)
            stampSaved := convertTimestampToNumber(stampSaved)
            stampDropped := convertTimestampToNumber(stampDropped)
            if( stampSaved > stampDropped ) {
              arrayOriginal[indexTranslated] := fileDropped
            }
          }else{
            arrayOriginal[indexTranslated] := fileDropped
          }
        }
      }
    }
  }
  
  for i, fileTranslated in arrayTranslated
  {
    Try fileOriginal := arrayOriginal[i]
    if ( fileOriginal == "" ) {
      continue
    }
    
    ;一時フォルダを中身ごと削除して再作成する
    Try DirDelete(exeDirOriginal, true)
    Try DirDelete(exeDirTranslated, true)
    Try DirDelete(exeDirOutput, true)
    
    DirCreate(exeDirOriginal)
    DirCreate(exeDirTranslated)
    
    ;フォルダ名とファイル名を取り出す
    SplitPath(fileTranslated, &fileNameTranslated, &dirTranslated)
    
    ;一時フォルダにコピーしたespからESLフラグを除去する
    filePathOriginal := exeDirOriginal . "\" . fileNameTranslated
    filePathTranslated := exeDirTranslated . "\" . fileNameTranslated
    FileCopy(fileOriginal, filePathOriginal, 1)
    FileCopy(fileTranslated, filePathTranslated, 1)
    removeEslFlag(filePathOriginal)
    removeEslFlag(filePathTranslated)

    if ( removeTranslatedPlugins != 0 ) {
      FileDelete(fileTranslated)
      FileMove(fileOriginal, fileTranslated)
    }

    ;Masterプラグインのリストを抽出する
    arrayMaster := extractMasterList(filePathTranslated)
    arrayMaster.Push(fileNameTranslated)
    
    ;esp2dsdを起動する
    RunWait('"' . exe . '" "' . exeDirOriginal . '" "' . exeDirTranslated . '"')
    
    ;espと同じ場所にjson生成用のディレクトリを作成する
    ;iniでenable_custom_output_folder_location=trueにしていると、代わりにoutput_folder_locationが使われる
    readConfig()
    if( enableCustomOutput != 0 ) {
      jsonDestDir := outputFolder . "\SKSE\Plugins\DynamicStringDistributor\" . fileNameTranslated . "\"
    }else{
      jsonDestDir := dirTranslated . "\SKSE\Plugins\DynamicStringDistributor\" . fileNameTranslated . "\"
    }
    DirCreate(jsonDestDir)
    
    Loop Files, exeDirOutput "\*.json"
    {
      ;json分割設定が無効な場合 (0.1.0までの挙動)
      if ( splitJsonByMaster == 0 ) {
        ;出力先指定ついでにjsonをリネームする
        fileNameOutput := RegExReplace(A_LoopFileName, "i)_output(\.es.\.json)$", "$1")

        ;古いファイルは削除しておく
        FileDelete(jsonDestDir . "\*.json")
        
        ;元のjsonを一行ずつ読み込んで処理する
        FileEncoding("UTF-8")
        Loop Read, exeDirOutput . "\" . A_LoopFileName, exeDirOutput . "\" . fileNameOutput
        {
          ;Masterリストを元にFormIDを正しい参照先で書き換える
          line := replaceFormID(A_LoopReadLine, arrayMaster)
          FileAppend(line . "`r`n")
        }
        FileCopy(exeDirOutput . "\" . fileNameOutput, jsonDestDir . "\" . fileNameOutput, 1)

      ;json分割設定が有効な場合 (0.2.0以降の挙動)
      }else{
        ;jsonファイルをMapオブジェクトに変換する
        textData := FileRead(exeDirOutput . "\" . A_LoopFileName)
        jsonObj := Jxon_Load(&textData)
        textData := ""

        ;古いファイルは削除しておく
        FileDelete(jsonDestDir . "\*.json")

        ;Masterごとに別の配列を作成する
        jsonMap := Map()
        for i, name in arrayMaster
        {
          jsonMap[name] := Array()
        }
        
        ;出力されたjsonをFormIDに修正を加えつつ振り分けていく
        for key, item in jsonObj
        {
          formId := item["form_id"]
          masterIdxHex := "0x" . SubStr(formId, 1, 2)
          masterIdxInt := Format("{1:u}", masterIdxHex + 0x1)
          masterName := arrayMaster[masterIdxInt]

          item["form_id"] := SubStr(item["form_id"], 1, 8) . "|" . masterName
          jsonMap[masterName].Push(item)
        }

        ;Masterごとにファイルに出力する
        for i, name in arrayMaster
        {
          if ( jsonMap[name].Length < 1) {
            continue
          }
          modifiedJson := Jxon_Dump(jsonMap[name], 4)
          if( name != fileNameTranslated ) {
            FileAppend(modifiedJson, jsonDestDir . "\" . fileNameTranslated . "___" . name . ".json")
          }else{
            FileAppend(modifiedJson, jsonDestDir . "\" . fileNameTranslated . ".json")
          }
        }
      }
    }
    
    ;結果表示用のカウンタと文字列を更新する
    countResult++
    strResult1 := strResult1 . "`n" . fileNameTranslated
  }
  
  ;結果表示
  if( countResult > 0 )
  {
    MsgBox(strResult0 . countResult . strResult1)
  }
}

;=====================================================
;  FormIDを正しい参照先で書き換える関数
;=====================================================

replaceFormID(str, arr) {
  for i, name in arr
  {
    idx := Format("{1:X}", i - 1)
    if (i < 0x10) {
      idx := "0" . idx
    }
    from := '"form_id": "' . idx . '([\da-fA-F]*)\|.*'
    to := '"form_id": "' . idx . '$1|' . name . '",'
    str := RegExReplace(str, from, to)
  }
  return str
}


;=====================================================
;  Masterの一覧を配列として取り出す関数
;=====================================================

extractMasterList(filePath) {
  fileObj := FileOpen(filePath, "r", "CP0") ;espをANSI形式で開く
  buf1 := Buffer(1)      ;バイト判定用のバッファ
  buf4 := Buffer(4)      ;ラベル判定用のバッファ
  arrayResult := Array() ;戻り値用の配列
  masterFound := 0       ;Master領域の終端判定用変数
  
  Loop
  {
    Loop
    {
      ;4バイトの覗き穴を1バイトずつ後ろにずらしながら読んでいく
      fileObj.RawRead(buf4, 4)
      
      ;Masterレコードのラベルを発見した場合
      if( StrGet(buf4, "CP65001") == "MAST" ) {
        
        ;MASTの1バイト前が0x00でなければポインタ位置を修正してループに戻る
        ;ポインタ移動は現在の覗き穴の末尾から次の覗き穴の先頭に向かって行う
        fileObj.Seek(-5, 1)
        fileObj.RawRead(buf1, 1)
        if( NumGet(buf1, "UChar") > 0 ) {
          fileObj.Seek(1, 1)
          continue
        }
        
        ;MASTの2バイト後が0x00でなければポインタ位置を修正してループに戻る
        fileObj.Seek(5, 1)
        fileObj.RawRead(buf1, 1)
        if( NumGet(buf1, "UChar") > 0 ) {
          fileObj.Seek(-5, 1)
          continue
        }
        
        ;プラグイン名の開始位置を記録して脱出
        masterStart := fileObj.Pos
        break
        
      ;既にMasterを発見しており、次の発見が途絶えた場合は配列を返して終了
      }else if( masterFound > 0 || fileObj.AtEOF > 0 ){
        fileObj.Close()
        return arrayResult
      
      ;ポインタを3バイト戻して次のループへ
      }else{
        fileObj.Seek(-3, 1)
      }
    }
    Loop
    {
      ;0x00が見えるまでポインタを進めて終了位置を記録
      fileObj.RawRead(buf1, 1)
      if( NumGet(buf1, "UChar") == 0 ) {
        masterEnd := fileObj.Pos
        break
      }
    }
    ;ポインタを開始位置まで戻し、終了位置まで覗き穴を広げる
    masterNameLength := masterEnd - masterStart
    fileObj.Seek(-masterNameLength, 1)
    
    ;文字列長のバッファを確保して読み出し、配列に保存する
    bufStr := Buffer(masterNameLength) 
    fileObj.RawRead(bufStr, masterNameLength)
    masterName := StrGet(bufStr, "CP65001")
    arrayResult.Push(masterName)
    
    ;終わったらバッファは開放
    bufStr := Buffer(0) 
    ;MsgBox("開始位置: " . masterStart . "`n終了位置: " . masterEnd . "`n文字列長: " . masterNameLength "`nマスター: " . masterName)
    
    ;ポインタを次のラベルまで14バイト進める
    fileObj.Seek(14, 1)
    
    ;最初のMasterが発見されたフラグ
    masterFound := 1
  }
}


;=====================================================
;  プラグインのESLフラグを除去する関数
;=====================================================

removeEslFlag(filePath) {
  fileObj := FileOpen(filePath, "rw", "CP0") ;espを読み書き可能な状態で開く
  buf1 := Buffer(1)        ;バイト判定用のバッファ
  fileObj.Seek(9, 1)       ;フラグがある位置まで移動
  fileObj.RawRead(buf1, 1) ;フラグを読み出す
  flags := NumGet(buf1, "UChar")   ;数値型に変換
  
  if( (flags & 0x2) > 0 ) {   ;ESLフラグを含んでいる場合
    flags := flags & (~0x2)   ;フラグを下ろしてから
    fileObj.Seek(-1, 1)       ;ポインタを1バイト戻して
    fileObj.WriteUChar(flags) ;編集済みのバイトで上書きする
  }
  fileObj.Close() ;ファイルハンドルを閉じる
}


;=====================================================
;  プラグインにESLフラグを付与する関数
;=====================================================


addEslFlag(filePath) {
  fileObj := FileOpen(filePath, "rw", "CP0") ;espを読み書き可能な状態で開く
  buf1 := Buffer(1)        ;バイト判定用のバッファ
  fileObj.Seek(9, 1)       ;フラグがある位置まで移動
  fileObj.RawRead(buf1, 1) ;フラグを読み出す
  flags := NumGet(buf1, "UChar")   ;数値型に変換
  
  if( (flags & 0x2) == 0 ) {  ;ESLフラグを含んでいない場合
    flags := flags | 0x2      ;フラグを立ててから
    fileObj.Seek(-1, 1)       ;ポインタを1バイト戻して
    fileObj.WriteUChar(flags) ;編集済みのバイトで上書きする
  }
  fileObj.Close() ;ファイルハンドルを閉じる
}


;=====================================================
;  ファイル名のタイムスタンプを数値に変換する関数
;=====================================================

convertTimestampToNumber(str) {
  out := StrReplace(str, "_")
  out := StrReplace(out, ".")
  return out
}


;=====================================================
;  iniを読み込む関数
;=====================================================

readConfig() {
  global iniPath
  global splitJsonByMaster
  global removeTranslatedPlugins
  global enableCustomOutput
  global outputFolder
  
  splitJsonByMaster := 0
  removeTranslatedPlugins := 0
  enableCustomOutput := 0
  outputFolder := "C:\Modding\MO2\overwrite"
  
  if !FileExist(iniPath)
  {
    FileAppend('[config]`r`n', iniPath)
    FileAppend('; jsonを上書きするプラグインごとに分割して保存するかどうか。`r`n', iniPath)
    FileAppend('; 0:無効, 1:有効`r`n', iniPath)
    FileAppend('split_json_by_master=0`r`n', iniPath)
    FileAppend('`r`n', iniPath)
    FileAppend('; 自動で翻訳済espを削除してバックアップを復元するかどうか。`r`n', iniPath)
    FileAppend('; 0:無効, 1:有効`r`n', iniPath)
    FileAppend('remove_translated_plugins=0`r`n', iniPath)
    FileAppend('`r`n', iniPath)
    FileAppend('; jsonの出力先を任意のフォルダに切り替えるかどうか。`r`n', iniPath)
    FileAppend('; 0:無効, 1:有効`r`n', iniPath)
    FileAppend('enable_custom_output_folder_location=0`r`n', iniPath)
    FileAppend('`r`n', iniPath)
    FileAppend('; 任意のフォルダのフルパス。この場所にSKSEフォルダが配置される。`r`n', iniPath)
    FileAppend('custom_output_folder_location=C:\Modding\MO2\overwrite`r`n', iniPath)
    
  }else{
    splitJsonByMaster := IniRead(iniPath, "config", "split_json_by_master", 0)
    removeTranslatedPlugins := IniRead(iniPath, "config", "remove_translated_plugins", 0)
    enableCustomOutput := IniRead(iniPath, "config", "enable_custom_output_folder_location", 0)
    outputFolder := IniRead(iniPath, "config", "custom_output_folder_location", "C:\Modding\MO2\overwrite")
  }
}


